// © Copyright Mocoolka Corporation 2015,2017.
// Node module: mocoolka-tools
// LICENSE: MIT

let typeTools = require('./type-tools.js');
let validTools = require('./validation-tools');
let iteratorTools = require('./iterator-tools');
let serviceTools = require('./service-tools');
/**
 * Error Tools static class
 *
 * Error :
 *  - ERROR/E-MISS-TYPE
 *  - ERROR/E-DIR-NOT-EXIST
 *  - ERROR/E-FILE-NOT-EXIST
 *  - ERROR/E-LANG-NOT-SUPPORT
 *  - ERROR/E-NOT-BLANK
 *  - ERROR/E-MISS-PARSE-FILE
 *  - ERROR/E-MISS-CALLBACK
 *  - ERROR/E-MISS-INPUT
 */
class ErrorTools {
  /**
   * validate json with schema,throw error if validate fail
   * @param {Object} msg
   * @propertyDetail {Object} data -actual json object
   * @propertyDetail {Object} schema -checked schema
   */
  static validateJsonSchema  (msg) {

    validTools.validateJsonSchema(msg, errors=> {

      if (errors) {

        let messages = [];
        iteratorTools.iterator(errors, (item)=> {
          let id = `VALIDATE/E-${item.keyword.toUpperCase()}`;
          let params = item.params;
          messages.push({ id, value: params });

        });

        ErrorTools.throwError({ id: 'ERROR/E-MISS-INPUT', value: messages });
      }

    });
  };

  /**
   * throw Error
   * @param {Object} error -param object
   * @propertyDetail {string} id - error id
   * @propertyDetail {Object} [value] -param value in intl
   * @propertyDetail {string} [errorLevel] -level :error ,warning ,default value is error
   * @propertyDetail {Object} [stack] - contain error stack
   */
  static throwError  (error)  {
    /*    ErrorTools.validateJsonSchema
        ({ schema: schemas.throwError, data: error });*/
    error.errorLevel = error.errorLevel || 'error';
    throw ErrorTools.getError(error);
  };

  /**
   * check error with id
   * @param {Object}msg -input object
   * @propertyDetail {Error} error - Error instance
   * @propertyDetail {string} id - expected error id
   * @returns {boolean} -true when error id equal expected error id,otherwise false
   */
  static checkError  (msg)  {

    ErrorTools.validateJsonSchema
    ({ schema: schemas.checkError, data: msg });
    return (msg.error.mkMessage && msg.error.mkMessage.id === msg.id);
  };

  /**
   * throw ERROR/E-MISS-PARSE-FILE error
   * @param {string}path - file path
   */

  static throwParseFileError  (path) {
    ErrorTools.throwError({
      id: 'ERROR/E-MISS-PARSE-FILE', value: {
        path,
      },
    });
  };

  static getError  (errorObject)  {
    let error = new Error(errorObject.id);
    error.mkMessage = errorObject;

    error.toString = function () {
      let result = `${this.mkMessage.id}:`;

      if (typeTools.isArray(this.mkMessage.value)) {
        this.mkMessage.value.map(item=> {
          result += `${item.id}: ${item.value ? `${JSON.stringify(item.value)}` : ''}
${item.options ? `options:${item.options}` : ''}
 `;
        });
      } else {
        result += ` ${this.mkMessage.value ? `${JSON.stringify(this.mkMessage.value)}` : ''}
${this.mkMessage.options ? `options:${this.mkMessage.options}` : ''} `;
      }

      return result;
    };

    return error;
  };

  /**
   * handler Error
   * @param {Object} error - Error instance
   * @param {Function} [callback]
   * @return {boolean} - true if the Error is Trusted; otherwise, false.
   */

  static errorHandler  (error)  {
    return new Promise(function (resolve, reject) {
      let messageContent;
      let isTrustedError = false;
      if (error.mkMessage) {
        isTrustedError = errorTools.trustedErrors.includes(error.mkMessage.errorLevel);
        error.mkMessage.errorLevel = error.mkMessage.errorLevel || 'warning';
        if (error.mkMessage.id === 'I18N.MISS-SYSTEM')
          messageContent = `${error.mkMessage.id}: miss system error message.`;
        else
          messageContent = error.toString();
        try {
          if (!error.mkMessage.stack) {
            error.mkMessage.stack = error.stack;
          }

          if (serviceTools.existService('mocoolka-log', 'rfc5424')) {
            serviceTools.standClient('mocoolka-log', 'rfc5424',
                {
                  level: error.mkMessage.errorLevel, message: error.mkMessage,
                }, (error, data)=> {
                  if (error)
                    console.error(error);
                });
            resolve({
              isTrustedError,
              message: messageContent, });
          } else {
            resolve({
              isTrustedError,
              message: messageContent, });
          }
        }
        catch (ex) {
          reject(ex);
        }

      } else {
        messageContent = error.toString();
        resolve({
          isTrustedError,
          message: messageContent, });
      }

    });
  };
}
const schemas = {
  throwError: {
    properties: {
      value: {
        type: 'object',
      },
      errorLevel: {
        type: 'string',
      },
      stack: {
        type: 'object',
      },
      id: {
        type: 'string',
      },
    },
    required: ['id'],
  },
  checkError: {
    properties: {
      error: {
        type: 'object',
      },
      id: {
        type: 'string',
      },
    },
    required: ['error', 'id'],
  },
  validateJsonSchema: {
    properties: {
      schema: {
        type: 'object',
      },
      data: {

      },
    },
    required: ['schema', 'data'],
  },
};
const errorTools = {
  trustedErrors: ['error', 'warning'],

};

module.exports.__proto__ = ErrorTools;
