// © Copyright Mocoolka Corporation 2015,2017.
// Node module: mocoolka-tools
// LICENSE: MIT

let mkDirP = require('mkdirp');
let pathSource = require('path');
let fs = require('fs');
let glob = require('glob').sync;
let errorTools = require('./error-tools.js');
const typeTools = require('./type-tools.js');
const TaskEmitter = require('./task/task-emitter');
const yml = require('yamljs');

/**
 * File Tools static class
 * @class FileTools
 */
class FileTools {
  /**
   * create directory when directory not exist
   * @param {string} path - full path
   */
  static createDirNotExist(path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.createDirNotExist, data: path });
    if (!FileTools.directoryExist(path)) {
      mkDirP.sync(path);
    }
  };

  /**
   * find file in directory using the patterns the shell uses, like stars and stuff.
   * @param {Object} msg -input object
   * @propertyDetail {string} path -search file in the directory
   * @propertyDetail {string} patterns  -file patterns . *.json
   * @return {Array} file array string
   */

  static findFiles  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.findFiles, data: msg });
    return glob(msg.patterns, { cwd: msg.path, nonull: false });
  };


  /**
   * search file in content array using extensions
   * @param {Object} msg - param object
   * @propertyDetail {string} path - search in the directory
   * @propertyDetail {string[]} content - content array.
   *  ['lib/*.js','test/*.json']
   * @propertyDetail {string[]} extensions - file extensions.['markdown','js']
   * @return {Promise}
   * @resolve
   * @param {Object[]} result - contain match files
   * @propertyDetail {String} filename -fileName
   * @propertyDetail {String} content -content
   */

  static matchFiles  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.matchFiles, data: msg });

    let root = msg.path;
    let content = msg.content;
    let extensions = msg.extensions;
    return new Promise(function (resolve, reject) {
      try {
        let files = {};
        let matchedFiles = [];

        content.forEach(function (p) {
          if (typeof p === 'string') {

            let matched = FileTools.findFiles({ path: root, patterns: p });
            matched.forEach(function (f) {
              if (matchedFiles.indexOf(f) === -1) {
                matchedFiles.push(f);
              }
            });
          } else {
            matchedFiles.push(p);
          }
        });

        matchedFiles.forEach(function (p) {
          if (typeof p === 'string') {
            let f = FileTools.path({ paths: [root, p] });

            if (extensions.includes(FileTools.getMKFileExtName(f))) {
              files[f] = FileTools.openFile({ path: f, mode: 'plain' });
            }
          }
        });

        let contents = [];
        matchedFiles.forEach(function (f) {

          if (typeof f === 'object') {
            contents.push({ content: f });
          } else {

            if (extensions.includes(FileTools.getMKFileExtName(f))) {
              f = FileTools.path({ paths: [root, f] });
              contents.push({ filename: f, content: files[f] });
            }
          }
        });

        resolve(contents);
      }
      catch (error) {
        reject(error);
      }

    });
  }

  /**
   * search file in directory using extensions.return all match filename
   * @param {Object} msg - param object
   * @propertyDetail {string} root - search in the directory
   * @propertyDetail {string[]} extensions - file extensions.etc['markdown','js']
   * @return {Promise}
   * @resolve {String[]} -file path array
   */

  static searchFiles  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.searchFiles, data: msg });

    let root = msg.path;
    let extensions = msg.extensions;
    return new Promise(function (resolve, reject) {
      let te = new TaskEmitter();
      let files = [];
      te.on('error', error=> {
        reject(error);
      });
      te.on('done', function () {
        resolve(files);
      });

      te.on('readdir', function (result) {
        let f = result.input[0];
        let dir = result.output[0];
        dir.forEach(function (df) {
          te.task({ scope: fs, name: 'stat', args: [FileTools.path({ paths: [f, df] })] });
        });
      });

      te.on('stat', function (result) {
        let f = result.input[0];
        let stat = result.output[0];
        if (stat.isDirectory()) {
          te.task({ scope: fs, name: 'readdir', args: [f] });
        } else {
          if (extensions.includes(FileTools.getMKFileExtName(f))) {
            files.push(f);
          }
        }
      });

      te.on('init', function iterateFiles() {
        te.task({ scope: fs, name: 'stat', args: [root] });
      });

      te.emit('init');
    });
  };

  /**
   * delete directory if directory exist and even directory contain files
   * @param {string} path
   */

  static deleteDirExist  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.deleteDirExist, data: path });
    if (FileTools.directoryExist(path)) {
      fs.readdirSync(path).forEach((file)=> {
        let curPath = path + '/' + file;
        if (fs.lstatSync(curPath).isDirectory()) { // recurse
          FileTools.deleteDirExist(curPath);
        } else { // delete file
          FileTools.deleteFile(curPath);
        }
      });

      fs.rmdirSync(path);
    }

  };

  /**
   * check file exist
   * @param {string} path -file path
   * @return {boolean}
   */

  static fileExist  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.fileExist, data: path });
    let fileStats;
    let validPath = true;
    try {
      fileStats = fs.statSync(path);
    } catch (e) {
      validPath = false;
    }

    if (validPath === false)
      return false;
    return (fileStats.isFile());
  };

  /**
   * check directory that is existed
   * @param {string} path -directory path
   * @return {boolean}
   */

  static directoryExist  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.directoryExist, data: path });
    let rootStats;
    let validPath = true;
    try {
      rootStats = fs.statSync(path);
    } catch (e) {
      validPath = false;
    }

    if (validPath === false)
      return false;
    return rootStats.isDirectory();

  };

  /**
   * Returns ext name of file name.
   * @param {string} path -file fullName that contain path
   * @return {string|null}
   */

  static getFileExtName  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.getFileExtName, data: path });
    let name = pathSource.parse(path).base;
    let parts = name.split('.');
    if (parts.length < 2) return null;
    return parts[parts.length - 1].toLowerCase();
  };

  /**
   * get directory from path
   * @param {string} path - full path
   * @returns {string}
   */
  static getDirectoryName  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.getDirectoryName, data: path });
    return pathSource.parse(path).dir;

  };

  /**
   * Returns  name of file name. the name is  name before first dot
   * @param {string} path -file fullName that contain path
   * @return {string|null}
   */

  static getMKFileName  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.getMKFileName, data: path });

    let name = pathSource.parse(path).base;
    if (typeof name !== 'string') return null;
    let parts = name.split('.');
    if (parts.length === 0) return null;
    return parts[0];
  };

  /**
   * Returns  name of file name.
   * @param {string} path file fullName that contain path
   * @return {string|null}
   */

  static getFileName  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.getFileName, data: path });

    return pathSource.parse(path).base;

  };
  /**
   * Returns  mk ext name of file name.
   * @param {string} path file fullName that contain path
   * @return {string|null}
   */

  static getMKFileExtName  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.getMKFileExtName, data: path });

    let name = pathSource.parse(path).base;
    let parts = name.split('.');
    if (parts.length < 2) return null;
    let extName = parts[1].toLowerCase();
    for (let i = 2; i < parts.length; i++) {
      extName += '.' + parts[i].toLowerCase();
    }

    return extName;
  };

  /**
   * open file as string or object
   * @param {Object} msg - input object
   * @propertyDetail {string} path - file path
   * @propertyDetail {string} [type] - support 'js' ,'json' 'yml' 'yaml'.default value is file extend name
   * @propertyDetail {string} [mode] - 'object' meaning open as object ,'string' meaning open as string,
   * default value is object
   * @propertyDetail {string} [encoding] - file encoding .default value is utf8
   *  - 'ascii'
   *  - 'utf8'
   *  - 'utf16le'
   *  - 'base64'
   *  - 'binary'
   *  - 'hex'
   * @returns {string|Object}
   */

  static openFile  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.openFile, data: msg });
    if (!FileTools.fileExist(msg.path))
      return null;
    let path = FileTools.realPath(msg.path);
    let type = msg.type || FileTools.getFileExtName(path);
    let mode = msg.mode || 'object';
    let result;
    if ((type === 'js' || type === 'json') && mode === 'object')
      result = require(path);
    else if ((type === 'yml' || type === 'yaml')  && mode === 'object')
      result = yml.parse(fileTools.openFileAsPlainString(msg));
    else
      result = FileTools.openFileAsPlainString(msg);
    return result;
  };

  /**
   * open file and return string
   * @param {Object} msg - input object
   * @propertyDetail {string} path - file path
   * @propertyDetail {string} [encoding] - file encoding .default value is utf8
   * @return {string} file content
   */

  static openFileAsPlainString  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.openFileAsPlainString, data: msg });
    let encoding = msg.encoding || 'utf8';
    return fs.readFileSync(msg.path, encoding);
  };

  /**
   * save file to path
   * @param {Object} msg - param object
   * @propertyDetail {string} path - save to the path
   * @propertyDetail {string|Object} content - file content will save to file
   */

  static saveFile  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.saveFile, data: msg });
    let path = msg.path;
    let content = msg.content;
    FileTools.createDirNotExist(FileTools.getDirectoryName(path));
    let type = FileTools.getFileExtName(path);
    if (type === 'json' && (typeTools.typeDetect(content) === 'Object'
      || typeTools.typeDetect(content) === 'Array'))
      content = JSON.stringify(content);
    fs.writeFileSync(path, content);
  };

  /**
   * delete file with file path
   * @param {string} path - file path
   */

  static deleteFile  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.deleteFile, data: path });
    fs.unlinkSync(path);
  };

  /**
   * return string that remove Bom
   * @param {string} str
   * @returns {string}
   */

  static stripBom  (str) {
    errorTools.validateJsonSchema
    ({ schema: schemas.stripBom, data: str });
    return str.charCodeAt(0) === 0xFEFF ? str.slice(1) : str;
  };

  /**
   * calculate path depth
   * @param {string} path - path
   * @returns {number}
   */

  static directoryDepth  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.directoryDepth, data: path });
    path = pathSource.normalize(path);
    return path.split(pathSource.sep).length;
  };

  /**
   * get real path
   * @param {string} path
   * @return {string|null}
   */

  static realPath  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.realPath, data: path });
    let realPath = null;
    try {
      realPath = process.browser ? path : fs.realpathSync(path);
    } catch (e) {
      return null;
    }

    return realPath;
  };

  /**
   * if path is symbol link ,return real path
   * @param {string} path
   * @returns {string|null}
   */

  static unsymbolLink  (path) {
    errorTools.validateJsonSchema
    ({ schema: schemas.unsymbolLink, data: path });
    if (!path) return null;
    let stats = null;
    try {
      stats = fs.lstatSync(path);
    } catch (e) {
      return null;
    }

    if (!stats) return null;
    if (stats.isSymbolicLink()) {
      let realPath = FileTools.realPath(path);

      return FileTools.unsymbolLink(realPath);
    } else {
      return stats.isDirectory() ? path : null;
    }
  };

  /**
   * get max directory depth
   * @returns {number}
   */

  static maxDirectoryDepth  () {
    let depth = parseInt(process.env.MOCOOLKA_GLOBALIZE_MAX_DEPTH, 10);
    if (isNaN(depth)) depth = FileTools.MAX_DIRECTORY_DEPTH;
    depth = Math.max(1, depth);
    return depth;
  };

  /**
   * if the depth that current directory in root directory
   * more than max directory depth return true
   * @param {string} msg -param object
   * @propertyDetail {string} path -current directory
   * @propertyDetail {string} root -root directory
   * @returns {boolean} - true meaning is exceed,otherwise false
   */

  static isExceedMaxDirectoryDepth  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.isExceedMaxDirectoryDepth, data: msg });

    let depthRoot = FileTools.directoryDepth(msg.root);

    let depthCurrent = FileTools.directoryDepth(msg.path);
    return (depthCurrent - depthRoot) > FileTools.maxDirectoryDepth();

  };

  /**
   *  joins all given path segments together using the platform specific separator
   *  as a delimiter, then normalizes the resulting path.
   * @param {Object} msg -param object

   * @property {string[]} paths - A array of path
   *
   *  path variable:
   * - $dirName = __dirname
   * - $process = process.cwd()
   * @propertyDetail {string} [dirName] - file path **__dirname**
   */

  static path  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.path, data: msg });
    let paths = msg.paths;
    let dirName = msg.dirName;
    let realPaths = [];

    if (typeTools.isArray(paths)) {
      paths.map(path=> {
        realPaths.push(pathResolve(path, dirName));
      });
    } else {
      realPaths.push(paths);
    }

    return pathSource.join.apply(null, realPaths);

  };

  /**
   * check path in root is full path
   * @param {Object} msg -param object
   * @propertyDetail {string} root -root path
   * @propertyDetail {string} path -check path in root path
   * @return {boolean} -true meaning is full path,otherwise false
   */

  static isFullPath (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.isFullPath, data: msg });
    console.log(pathSource.resolve(msg.root, msg.path));
    return pathSource.resolve(msg.root, msg.path) === msg.path;
  };
}
const pathResolve = (path, dirName)=> {

  if (path === '$dirName') {
    return dirName;
  } else if (path === '$process') {
    return process.cwd();
  } else
    return path;
};

const schemas = {
  isFullPath: {
    type: 'object',
    properties: {
      root: {
        type: 'string',
      },
      path: {
        type: 'string',
      },
    },
    required: ['path', 'root'],
  },
  path: {
    type: 'object',
    properties: {
      paths: {
        type: 'array',
        items: { type: 'string' },
      },
      dirName: {
        type: 'string',
      },
    },
    required: ['paths'],
  },
  isExceedMaxDirectoryDepth: {
    type: 'object',
    properties: {
      root: {
        type: 'string',
      },
      path: {
        type: 'string',
      },
    },
    required: ['path', 'root'],
  },
  findFiles: {
    type: 'object',
    properties: {
      path: {
        type: 'string',
      },
      patterns: {
        type: 'string',
      },
    },
    required: ['path', 'patterns'],
  },
  createDirNotExist: {
    type: 'string',
  },
  deleteDirExist: {
    type: 'string',
  },
  fileExist: {
    type: 'string',
  },
  directoryExist: {
    type: 'string',
  },
  getFileExtName: {
    type: 'string',
  },
  getDirectoryName: {
    type: 'string',
  },
  getMKFileName: {
    type: 'string',
  },
  getFileName: {
    type: 'string',
  },
  getMKFileExtName: {
    type: 'string',
  },
  deleteFile: {
    type: 'string',
  },
  stripBom: {
    type: 'string',
  },
  directoryDepth: {
    type: 'string',
  },
  realPath: {
    type: 'string',
  },
  unsymbolLink: {
    type: 'string',
  },
  saveFile: {
    type: 'object',
    properties: {
      path: {
        type: 'string',
      },
      content: {
        anyOf: [
          { type: 'string' },
          { type: 'object' },
        ],
      },
    },
    required: ['path', 'content'],
  },
  matchFiles: {
    type: 'object',
    properties: {
      path: {
        type: 'string',
      },
      content: {
        type: 'array',
        items: {
          anyOf:[
            { type: 'string' },
            { type: 'object' },
          ],
        },
      },
      extensions: {
        type: 'array',
        items: { type: 'string' },
      },
    },
    required: ['path', 'content', 'extensions'],
  },
  searchFiles: {
    type: 'object',
    properties: {
      path: {
        type: 'string',
      },
      extensions: {
        type: 'array',
        items: { type: 'string' },
      },
    },
    required: ['path', 'extensions'],
  },
  openFile: {
    type: 'object',
    properties: {
      path: {
        type: 'string',
      },
      type: {
        type: 'string',
      },
      mode: {
        type: 'string',
      },
      encoding: {
        type: 'string',
      },
    },
    required: ['path'],
  },
  openFileAsPlainString: {
    type: 'object',
    properties: {
      path: {
        type: 'string',
      },
      encoding: {
        type: 'string',
      },
    },
    required: ['path'],
  },
};
FileTools.MAX_DIRECTORY_DEPTH = 99;
module.exports.__proto__ = FileTools;

