// © Copyright Mocoolka Corporation 2015,2017.
// Node module: mocoolka-tools
// LICENSE: MIT

const  typeTools    = require('./type-tools.js');
const  iteratorTools    = require('./iterator-tools.js');
const errorTools= require('./error-tools.js');
const deep = require('deep-diff');

/**
 * Object static tools class
 * @class
 */
class ObjectTools {

  /**
   * deep clone object
   * @param {Object} input - the object will clone
   * @returns {Object} cloned object
   */
  static clone  (input) {
    errorTools.validateJsonSchema
    ({ schema: schemas.clone, data: input });
    let output = input;
    let type = typeTools.typeDetect(input);
    if (type === typeTools.TYPES.Array) {
      output = [];
      let size = input.length;
      input.map(item=> {
        output.push(ObjectTools.clone(item));
      });
    } else if (type === typeTools.TYPES.Object) {
      output = {};
      for (let index in input) {
        output[index] = ObjectTools.clone(input[index]);
      }
    }

    return output;
  };



  /**
   * merge all input using clone and recursive
   * @param {Object[]} objects -the values will merge to result
   * @return {Object} merged object
   */

  static merge(objects) {
    errorTools.validateJsonSchema
    ({ schema: schemas.merge, data: objects });
    return mergeBase(true, true, objects);
  };

  /**
   * merge all input object.
   * @param {Object} msg -input message
   * @propertyDetail {boolean} isClone -if true use clone
   * @propertyDetail {boolean} isRecursive - if true use recursive.
   * @propertyDetail {Object[]} objects -The Objects will merge to result
   * @return {Object} merged object
   */
  static mergeAll(msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.mergeAll, data: msg });
    return mergeBase(msg.isClone, msg.isRecursive, msg.objects);
  };

  /**
   * set object value with alias value,remove alias property
   * @param {Object} msg
   * @propertyDetail {Object} obj -The Object contain values
   * @propertyDetail {Object} alias - The Object key is source property ,value is alias.value type must is string
   * or [string]
   * @return {Object} merged object with alias
   */
  static alias(msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.alias, data: msg });
    if (msg && msg.obj && msg.alias) {
      iteratorTools.iterator(msg.alias, (item)=> {
        let alias = item.value;
        if (typeTools.isString(alias)) {
          alias = [alias];
        }

        iteratorTools.iterator(alias, (aliasValue)=> {
          if (msg.obj[aliasValue]) {
            msg.obj[item.key] = msg.obj[aliasValue];
            delete msg.obj[aliasValue];
          }
        });

      });
      return msg.obj;
    }
  }

  /**
   * calculates the differences between two objects,
   * @param {Object} msg -contain source and dest
   * @propertyDetail {Object} source - source object
   * @propertyDetail {Object} dest - dest object
   * @return {Object[]} difference Object Array
   * @propertyDetail {string} kind -indicates the kind of change; will be one of the following:
   *
   * * `N` - indicates a newly added property/element
   * * `D` - indicates a property/element was deleted
   * * `E` - indicates a property/element was edited
   * * `A` - indicates a change occurred within an array
   * @propertyDetail {Array} path -the property path (from the left-hand-side root)
   * @propertyDetail {string} lhs - the value on the left-hand-side of the comparison (undefined if kind === 'N')
   * @propertyDetail {string} rhs -the value on the right-hand-side of the comparison (undefined if kind === 'D')
   * @propertyDetail {number} index -when kind === 'A',
   * indicates the array index where the change occurred
   * @propertyDetail {Object} item -when kind === 'A',
   * contains a nested change record indicating the change that occurred at the array index
   */
  static diff(msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.diff, data: msg });
    return ObjectTools.clone(deep.diff(msg.source, msg.dest));
  };
}
/**
 * Merge two objects recursively
 * @private
 * @param {Object} base -input
 * @param {Object} extend -extend
 * @return {Object} merge object with inputs
 */
const  mergeRecursiveBasic = (base, extend)=> {
  if (!typeTools.isObject(base))
    return extend;

  for (let key in extend) {

    if (typeTools.isObject(base[key]) && typeTools.isObject(extend[key])) {

      base[key] = mergeRecursiveBasic(base[key], extend[key]);

    } else {
      base[key] = extend[key];
    }

  }

  return base;
};
/**
 * Merge two or more objects
 * @private
 * @param {boolean} isClone -clone when true
 * @param {boolean} isRecursive -recursive when value is true
 * @param {Array} args
 * @return object
 */
const mergeBase =  (isClone, isRecursive, args)=> {
  if (args.length == 0)
    return {};
  let result = ObjectTools.clone(args[0]);
  let size = args.length;

  if (isClone || !typeTools.isObject(result))
    result = {};

  for (let index = 0; index < size; ++index) {
    let item = args[index];
    if (!typeTools.isObject(item))
      continue;

    for (let key in item) {
      let sitem = isClone ? ObjectTools.clone(item[key]) : item[key];
      if (isRecursive) {
        result[key] = mergeRecursiveBasic(result[key], sitem);
      } else {
        result[key] = sitem;
      }
    }
  }

  return result;

};

const schemas = {
  clone: {

  },
  merge: {
    type: 'array',
    items: { type: 'object' },
  },
  mergeAll: {
    type: 'object',
    properties: {
      isClone: {
        type: 'boolean',
        default: false,
      },
      isRecursive: {
        type: 'boolean',
        default: false,
      },
      objects: {
        type: 'array',
        items: { type: 'object' },
      },
    },
    required: ['objects'],
  },
  alias: {
    type: 'object',
    properties: {
      obj: {
        type: 'object',
      },
      alias: {
        type: 'object',
      },
    },
    required: ['obj', 'alias'],
  },
  diff: {
    type: 'object',
    properties: {
      source: {
        type: 'object',
      },
      dest: {
        type: 'object',
      },
    },
    required: ['source', 'dest'],
  },

};
module.exports = ObjectTools;
