// © Copyright Mocoolka Corporation 2015,2017.
// Node module: mocoolka-tools
// LICENSE: MIT

const iteratorTools = require('./iterator-tools');
const seneca = require('seneca');
const errorTools = require('./error-tools');
/**
 * service static class
 * @class
 */
class ServiceTools {
  /**
   * add action in service.
   * @private
   * @param {Object} service -add to the service object
   * @param {string} moduleName -request module name
   * @param {string} actionName -request action name
   * @param {Function} action -action function
   * @param {boolean} isPromise -Is action function **Promise**
   */
  static addAction  (service, moduleName, actionName, action, isPromise) {
    if (isPromise) {
      service.add({ role: moduleName, cmd: actionName }, function (msg, callback) {
        try {
          action(msg).then(result=> {
            callback(null, result);
          }).catch(error=> {
            callback(error);
          });
        }
        catch (error) {
          callback(error);
        }
      });
    } else {
      service.add({ role: moduleName, cmd: actionName }, action);
    }
  };

  /**
   * **seneca** plugin
   * @private
   * @param {Object} options
   * @propertyDetail {string} moduleName -request module name
   * @propertyDetail {Function} initPromise -module init function using **Promise**
   * @propertyDetail {Function} init -module init function using **callback**
   * @propertyDetail {Object[]} actionPromises - service action options using **Promise**
   * @propertyDetail {Function[]} actions - service action options using **callback**
   */
  static standPlugin(options) {
    this.module = options;
    this.initActions = function (callback) {
      //execute action callback
      if (this.module.actions) {
        iteratorTools.iterator(this.module.actions, (obj)=> {
            ServiceTools.addAction(this, this.module.moduleName, obj.key, obj.value, false);
          });
      }

      //execute action Promise
      if (this.module.actionPromises) {
        iteratorTools.iterator(this.module.actionPromises, (obj)=> {
          ServiceTools.addAction(this, this.module.moduleName, obj.key, obj.value, true);
        });
      }

      callback(null);
    };

    this.add('init:standPlugin', function (initOptions, callback) {
      if (this.module.init) {
        ServiceTools.addAction(this, this.module.moduleName, 'init', this.module.init, false);
        this.act({ role: this.module.moduleName, cmd: 'init' },
          this.module.initOptions, function (error) {
          if (error)
            callback(error);
          this.initActions(callback);
        });
      } else if (this.module.initPromise) {
        ServiceTools.addAction(this, this.module.moduleName, 'initPromise',
          this.module.initPromise, true);
        this.act({ role: this.module.moduleName, cmd: 'initPromise' },
          this.module.initOptions, function (error) {
          if (error)
            callback(error);
          this.initActions(callback);
        });
      } else {
        this.initActions(callback);
      }
    });

  }

  /**
   * start services
   * @param {Object} msg
   * @propertyDetail {string} moduleName -request module name
   * @propertyDetail {Function} [initPromise] -module init function using **Promise**
   * @propertyDetail {Function} [init] -module init function using **callback**
   * @propertyDetail {Function[]} [actionPromises] - service action options using **Promise**
   * @propertyDetail {Function[]} [actions] - service action options using **callback**
   * @propertyDetail {Object} [initOptions] - service init options
   * @propertyDetail {Object} service - service network config
   * @propertyDetail {Object} [service.transport] - service transport
   * @propertyDetail {Object} service.listen - service listen
   * @return  {Promise}
   */
  static standServices(msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.standServices, data: msg });
    return new Promise(function (resolve, reject) {
      try {
        let service = seneca({ log: 'silent' });
        let count = 0;
        let readyCount = 0;

        let transport = msg.service.transport;
        if (transport) {
          service = service.use(transport);
        }

        service = service.use(ServiceTools.standPlugin, msg);

        service.listen(msg.service.listen)
          .ready(function () {
              resolve();
            });

      }
      catch (error) {
        reject(error);
      }

    });

  }

  /**
   * execute action for service with callback
   * @param {Object} msg
   * @propertyDetail {string} moduleName - module name
   * @propertyDetail {string} actionName - action name
   * @propertyDetail {Object} message - action params
   * @propertyDetail {Object} service - service network config
   * @propertyDetail {Object} [service.transport] - service transport
   * @propertyDetail {Object} service.listen - service listen
   * @callback callback The callback return result
   * @param {Error} error -return error when execute the method.return null when no error
   * @param {*} result -return action result
   */
  static standClient(msg, callback) {
    errorTools.validateJsonSchema
    ({ schema: schemas.standClientPromise, data: msg });
    ServiceTools.buildServiceClient(msg.service).
    act({ role: msg.moduleName, cmd: msg.actionName }, msg.message, callback);
  };

  static buildServiceClient(msg) {
    let service = seneca({ log: 'silent' });
    if (msg.transport) {
      service = service.use(msg.transport);
    }

    return service.client(msg.listen);
  }

  /**
   * execute client action for service as promise
   * @param {Object} msg
   * @propertyDetail {string} moduleName - module name
   * @propertyDetail {string} actionName - action name
   * @propertyDetail {Object} message - action params
   * @propertyDetail {Object} service - service network config
   * @propertyDetail {Object} [service.transport] - service transport
   * @propertyDetail {Object} service.listen - service listen
   * @return {Promise}
   * @resolve
   * @param {*} result -return action result
   */
  static standClientPromise  (msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.standClientPromise, data: msg });
    return new Promise(function (resolve, reject) {

      ServiceTools.buildServiceClient(msg.service).
      act({ role: msg.moduleName, cmd: msg.actionName },
        msg.message, function (error, data) {
        if (error)
          reject(error);
        resolve(data);
      });
    });
  };

  /**
   * check service using moduleName and actionName
   * @param {Object} msg
   * @propertyDetail {string} moduleName - module name
   * @propertyDetail {string} actionName - action name
   * @propertyDetail {Object} service - service network config
   * @propertyDetail {Object} [service.transport] - service transport
   * @propertyDetail {Object} service.listen - service listen
   * @returns {boolean}
   */
  static existService(msg) {
    errorTools.validateJsonSchema
    ({ schema: schemas.existService, data: msg });
    return ServiceTools.buildServiceClient(msg.service).
    has({ role: msg.moduleName, cmd: msg.actionName });
  };

}
const schemas = {
  standServices: {
    type: 'object',
    properties: {
      moduleName: {
        type: 'string',
      },
    },
    required: ['moduleName'],
  },
  standClientPromise: {
    type: 'object',
    properties: {
      service: {
        type: 'object',
        properties: {
          transport: { type: 'object' },
          listen: { type: 'object' },
        },
        required: ['listen'],
      },
      moduleName: {
        type: 'string',
      },
      actionName: {
        type: 'string',
      },
      message: {
        type: 'object',
      },
    },
    required: ['service', 'moduleName', 'actionName','message'],
  },
  existService: {
    type: 'object',
    properties: {
      service: {
        type: 'object',
        properties: {
          transport: { type: 'object' },
          listen: { type: 'object' },
        },
        required: ['listen'],
      },
      moduleName: {
        type: 'string',
      },
      actionName: {
        type: 'string',
      },
    },
    required: ['service', 'moduleName', 'actionName'],
  },
};
module.exports = ServiceTools;
