// © Copyright Mocoolka Corporation 2015,2017.
// Node module: mocoolka-tools
// LICENSE: MIT

const md5Source = require('md5');
const memoize = require('lodash/memoize');
const util = require('util');
const _string = require('underscore.string');
/**
 * String tools static class
 * @class
 */
class StringTools{

  /**
   * convert message using md5
   * @param {string|buffer|Array} message -the message will md5
   * @param {Object} [options] -encoding options ,default is Hex
   * @propertyDetail {boolean} [asBytes] -return binary encoding
   * @propertyDetail {boolean} [asString] -return binary to string
   * @return {string|binary}
   */
  static md5  (message, options) {
    return md5Source(message, options);
  };

  /**
   * cache md5 result value
   * @param {string|buffer|Array}message
   * @param {Object}options encoding:'binary' message type is binary
   * asBytes:true return binary,asString:true return binary to string
   * @return {string|binary}
   */
  static md5Cache  (message, options) {
    return (memoize(md5))(message, options);
  };

  /**
   * format message using variable
   * @param {...string|number|json} args
   * @return {string}
   */
  static format  (...args) {
    return util.format.apply(null, args);
  };

  /**
   * Converts underscored or dasherized string to a camelized one.
   * Begins with a lower case .
   * @param {string} str
   * @return {string|*} convert result value
   */
  static camelize  (str) {
    str = _string.camelize(str);
    str = str.slice(0, 1).toLowerCase() + str.slice(1);
    return str;
  };
  /**
   * Converts string to camelized class name. First letter is always upper case
   * @param {string} str the value be converted
   * @return {string} convert result value
   */
  static classify  (str) {
    return _string.classify(str);
  };
  /**
   * Transform text into an ascii slug which can be used in safely in URLs.
   * Replaces whitespaces, accentuated, and special characters with a dash.
   * Limited set of non-ascii characters are transformed to similar versions in the ascii character set such as ä to a.
   * @param {string} str the value be converted to url
   * @returns {string} safe url string
   */
  static urlSafe  (str) {
    return _string.slugify(str.toLowerCase().replace(/\./g, '-'));
  };

  /**
   * Converts a camelized or dasherized string into an underscored one
   * @param {string} str -the value be converted
   * @return {string} convert result value
   */
  static underscored  (str) {
    return _string.underscored(str);
  };
  /**
   * Converts a underscored or camelized string into an dasherized one
   * @param {string} str the value be converted
   * @return {string} convert result value
   */
  static dasherize  (str) {
    return _string.dasherize(str);
  };
}

module.exports = StringTools;
